#! /usr/bin/env python2
#
# hcron_conv.py

# GPL--start
# This file is part of hcron
# Copyright (C) 2008-2019 Environment/Environnement Canada
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# GPL--end

"""Convert between crontab and hcron formats.
"""

# system imports
import os
import os.path
import socket
import sys
from sys import stderr
import traceback

# app import
from hcron.constants import *
from hcron.event import Event

def convert_to_events_main(args):
    """main for "--to-events".
    """
    try:
        crontabpath = None
        dirpath = None
        hostname = None
        mailaddr = ""

        while args:
            arg = args.pop(0)
            if arg in [ "--mail" ]:
                mailaddr = args.pop(0)
            elif len(args) == 2:
                hostname, crontabpath, dirpath = [arg]+args
                del args[:]
            else:
                raise Exception()

        if None in [crontabpath, dirpath, hostname]:
            raise Exception()

    except:
        #traceback.print_exc()
        raise

    try:
        convert_to_events(hostname, crontabpath, dirpath, mailaddr=mailaddr)
    except:
        pass

def convert_to_events(hostname, crontabpath, dirpath, mailaddr=""):
    try:
        if crontabpath == "-":
            crontabfile = sys.stdin
        else:
            crontabfile = open(crontabpath, "r")
    except Exception:
        stderr.write("error: could not open crontab file (%s)\n" % crontabpath)
        sys.exit(1)

    # generate and store each event
    events = []
    try:
        l = [ "# generated by hcron-conv" ]
        for line in crontabfile:
            line = line.strip()

            if line == "":
                continue

            if line.startswith("#"):
                l.append(line)
                continue
            elif line.startswith("@"):
                alias, cmd = line.split(None, 1)
                if alias in CRONTAB_ALIASES_MAP:
                    minute, hour, dom, month, dow = CRONTAB_ALIASES_MAP[alias].split(None, 4)
                else:
                    # unsupported alias
                    continue
            elif not line[0].isdigit():
                # variable
                continue
            else:
                minute, hour, dom, month, dow, cmd = line.split(None, 5)
            if month.isalpha():
                monthname = month.lower()[:3]
                month = MONTH_NAMES_MAP.get(monthname)
            if dow.isalpha():
                downame = dow.lower()[:3]
                dow = DOW_NAMES_MAP.get(downame)

            if month == None or dow == None:
                # error in spec
                continue

            d = HCRON_EVENT_DEFINITION_MAP.copy()
            d["host"] = hostname
            d["command"] = cmd
            d["notify_email"] = mailaddr
            d["when_month"] = month
            d["when_day"] = dom
            d["when_hour"] = hour
            d["when_minute"] = minute
            d["when_dow"] = dow

            st = "\n".join([ "%s=%s" % (name, d[name]) for name in HCRON_EVENT_DEFINITION_NAMES ])

            l.append(st)
            events.append("\n".join(l))
            l = ["# generated by hcron-conv"]
    except Exception:
        stderr.write("error: could not process crontab file (%s)\n" % crontabpath)
        sys.exit(1)

    crontabfile.close()

    # make sure the dir exists
    try:
        if not os.path.isdir(dirpath):
            os.makedirs(dirpath)
    except Exception:
        stderr.write("error: could not create directory (%s)\n" % dirpath)
        sys.exit(1)

    # write the events to individual event definition files
    for num, event in zip(range(len(events)), events):
        path = os.path.join(dirpath, str(num))
        try:
            open(path, "w+").write(event)
        except Exception:
            stderr.write("error: could not create event definition file (%s)\n" % path)
            sys.exit(1)

def convert_to_crontab_main(args):
    """main for "--to-crontab".
    """
    try:
        crontabpath = None
        dirpath = None
        remoteshell = "ssh"

        while args:
            arg = args.pop(0)
            if arg in ["--remoteshell", "--remoteshell"]:
                remoteshell = args.pop(0)
            elif len(args) == 1:
                crontabpath, dirpath = [arg]+args
                del args[:]
            else:
                raise Exception()

        if None in [crontabpath, dirpath]:
            raise Exception()
    except:
        #traceback.print_exc()
        raise

    try:
        convert_to_crontab(crontabpath, dirpath, remoteshell=remoteshell)
    except:
        pass

def convert_to_crontab(crontabpath, dirpath, remoteshell):
    try:
        if crontabpath == "-":
            crontabfile = sys.stdout
        else:
            crontabfile = open(crontabpath, "w+")
    except Exception:
        stderr.write("error: could not open crontab file (%s)\n" % crontabpath)
        sys.exit(1)

    # collect all events in dir tree
    l = ["# generated by hcron-conv"]
    for root, dirnames, filenames in os.walk(dirpath):
        for filename in filenames:
            path = os.path.join(root, filename)
            event = Event("", "", autoload=False)
            event.load(path)
            d = dict(event.assignments)

            l.append("# %s" % path)
            st = "%(when_minute)s %(when_hour)s %(when_day)s %(when_month)s %(when_dow)s" % d
            if d.get("host") in ["", None]:
                st2 = ""
            else:
                if d.get("as_user"):
                    st2 = "%s %s@%s" % (remoteshell, d["as_user"], d["host"])
                else:
                    st2 = "%s %s" % (remoteshell, d["host"])
            if d.get("command"):
                l.append("%s %s %s" % (st, st2, d["command"]))
            else:
                l.append("# skipped; no command")

    # write crontab file
    crontabfile.write("\n".join(l))
    crontabfile.close()

def print_usage():
    print("""\
usage: hcron conv --to-events [<options>] <host> <crontabpath> <dirpath>
       hcron conv --to-crontab [<options>] <crontabpath> <dirpath>
       hcron conv -h|--help

Convert between crontab and hcron event file formats.

Use --to-events to convert from a basic crontab to hcron event files.
An event file is created for each crontab command. Event files are
put into a directory.

Use --to-crontab to convert hcron event files to a crontab. A crontab
command is created for each event file. Event files are taken from a
directory.

Options for --to-events:
--mail <address>    Email address used when populating the "mail"
                    setting.

Options for --to-crontab:
--remoteshell <shell>
                    Remote shell to prepend to each crontab command when
                    the "host" setting is not empty. Default is ssh.""")

def main(args):
    try:
        arg = args.pop(0)
        if arg == "--to-events":
            convert_to_events_main(args)
        elif arg == "--to-crontab":
            convert_to_crontab_main(args)
        elif arg in ["-h", "--help"]:
            print_usage()
            sys.exit(0)
        else:
            raise Exception()
    except SystemExit:
        raise
    except Exception:
        #traceback.print_exc()
        stderr.write("error: bad/missing argument\n")
        sys.exit(1)

    sys.exit(0)
